#include "jannex/jannex.h"

#include "jannex/ota.h"
#include "jannex/api/console.h"
#include "jannex/api/coredump.h"

#include <string.h>
#include "esp_log.h"
#include "esp_ota_ops.h"
#include "esp_event.h"
#include "nvs_flash.h"

#include <stdio.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#include "esp_partition.h"
#include "esp_netif.h"
#include "esp_netif_sntp.h"
#include "esp_sntp.h"

#include "esp_private/panic_internal.h"
#include "esp_debug_helpers.h"
#include "esp_chip_info.h"

extern uint16_t panicRetryCounter;    // RTC_NOINIT_ATTR
extern uint16_t softwareResetCounter; // RTC_NOINIT_ATTR

const char *reset_reason_to_string(esp_reset_reason_t reason);
const char *getChipModelString(esp_chip_model_t chip_model);

void time_sync_notification_cb(struct timeval *tv)
{
    ESP_LOGI(jannexTAG, "Notification of a time synchronization event");
}

void initialize_sntp()
{
    ESP_LOGI(jannexTAG, "Initializing SNTP...");

    // set timezone to berlin
    // setenv("TZ", "CET-1CEST,M3.5.0,M10.5.0/3", 1);
    // tzset();

    esp_sntp_config_t config = {
        .start = true,
        .sync_cb = time_sync_notification_cb,
        .smooth_sync = false,
        .server_from_dhcp = true,           // accept NTP offers from DHCP server, if any (need to enable *before* connecting)
        .renew_servers_after_new_IP = true, // let esp-netif update configured SNTP server(s) after receiving DHCP lease
        .index_of_first_server = 1,         // updates from server num 1, leaving server 0 (from DHCP) intact
        .num_of_servers = 1,                // update only one server
        .servers = {CONFIG_NTP_SERVER},     // server pool

    };

    /*esp_netif_sntp_init(&config);

    esp_netif_sntp_start();

    // get current time in human readable format and print to log
    time_t now;
    struct tm timeinfo;
    time(&now);
    localtime_r(&now, &timeinfo);

    if (!localtime(&timeinfo))
    {
        ESP_LOGE(jannexTAG, "Failed to obtain time");
    }
    else
    {
        ESP_LOGI(jannexTAG, "The current date/time in Berlin is: %s", asctime(&timeinfo));
    }*/
}

void startJannex(void *parameter)
{
    defaultValuesToStorage();

#ifdef WIFI_ENABLED
    esp_log_level_set("esp-x509-crt-bundle", ESP_LOG_WARN);
    esp_log_level_set("wifi", ESP_LOG_WARN);
    init_wifi();
#endif

#ifdef WEB_SERVER_ENABLED
    initServer();
#endif

#ifdef CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE
    connectConsole();
#endif // CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE

    initialize_sntp();

    vTaskDelete(NULL);
}

void initJannex()
{
    initStorage();
    ESP_ERROR_CHECK(esp_netif_init());
    ESP_ERROR_CHECK(esp_event_loop_create_default());
#ifdef CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE
    initConsole();
#endif // CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE

    const esp_app_desc_t *app_desc = esp_app_get_description();

    esp_chip_info_t chip_info;
    esp_chip_info(&chip_info);

    ESP_LOGI(jannexTAG, "Initializing Jannex");

    bool coredumpAvailable = false;

    // check for core dump from flash partition
    esp_partition_t *coredump_partition = esp_partition_find_first(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_DATA_COREDUMP, "coredump");

    uint8_t *buffer = (uint8_t *)malloc(coredump_partition->size);
    esp_err_t err = esp_partition_read(coredump_partition, 0, buffer, coredump_partition->size);

    if (err == ESP_OK)
    {
        bool cdAvailable = false;
        // sum the first 100 bytes of buffer
        for (int i = 0; i < 3000; i++)
        {
            if (buffer[i] != 0xFF)
            {
                cdAvailable = true;
                break;
            }
        }

        if (cdAvailable)
        {
            coredumpAvailable = true;
            ESP_LOGE(jannexTAG, "---------- CORE DUMP INFO ----------");
        }
    }
    else
    {
        ESP_LOGW(jannexTAG, "Coredump part read failed: %s", esp_err_to_name(err));
    }
    free(buffer);

    esp_reset_reason_t reset_reason = esp_reset_reason();

    ESP_LOGI(jannexTAG, "JNXJSON{\n"
                        "  \"ver\": \"%s\",\n"
                        "  \"chip\": \"%s\",\n"
                        "  \"chipRev\": \"%i\",\n"
                        "  \"name\": \"%s\",\n"
                        "  \"build_date\": \"%s\",\n"
                        "  \"build_time\": \"%s\",\n"
                        "  \"rstRsn\": \"%s\",\n"
                        "  \"softRst\": %i,\n"
                        "  \"panicRst\": %i\n"
                        "}JNXJSON",
             app_desc->version,
             getChipModelString(chip_info.model),
             chip_info.revision,
             app_desc->project_name,
             app_desc->date,
             app_desc->time,
             reset_reason_to_string(reset_reason),
             softwareResetCounter,
             panicRetryCounter);

    /*ESP_LOGI(jannexTAG, "Reset: %s", reset_reason_to_string(reset_reason));

    ESP_LOGI(jannexTAG, "Software resets: %i", softwareResetCounter);
    ESP_LOGI(jannexTAG, "Panic retries: %i", panicRetryCounter);

    ESP_LOGI(jannexTAG, "Build version: %s", app_desc->version);
    ESP_LOGI(jannexTAG, "\" name: %s", app_desc->project_name);
    ESP_LOGI(jannexTAG, "\" date: %s", app_desc->date);
    ESP_LOGI(jannexTAG, "\" time: %s", app_desc->time);

    ESP_LOGI(jannexTAG, "Chip model: %s", getChipModelString(chip_info.model));
    ESP_LOGI(jannexTAG, "\" revision: %i", chip_info.revision);*/

    if (reset_reason == ESP_RST_POWERON)
    {
        softwareResetCounter = 0;
        panicRetryCounter = 0;
    }
    else
    {
        softwareResetCounter++;
    }

    if (coredumpAvailable)
    {

        ESP_LOGE(jannexTAG, "-------- CORE DUMP INFO END --------");
#ifdef CONFIG_DEBUG_SEND_COREDUMP
        ESP_LOGI(jannexTAG, "Core dump available. Sending it to the server:");
        sendCoredumpToConsole();
#else
        ESP_LOGI(jannexTAG, "Core dump available. Please download it from the web interface.");
#endif
    }

    xTaskCreate(startJannex, "startJannex", 4096, NULL, 5, NULL);
    xTaskCreate(checkIfOTAwasSuccessful, "checkIfOTAwasSuccessful", 4096, NULL, 5, NULL);
}

void updateJannex()
{
    xTaskCreate(checkForUpdatesAndInstall, "checkForUpdatesAndInstall", 8192, NULL, 5, NULL);
}

// Declare the real panic handler function. We'll be able to call it after executing our custom code
void __real_esp_panic_handler(void *);

// This function will be considered the esp_panic_handler to call in case a panic occurs
void __wrap_esp_panic_handler(panic_info_t *info)
{
    // Custom code, count the number of panics or simply print a message
    panicRetryCounter++;

    // Convert the integer to a string
    char retryCounterStr[10]; // Adjust the size as needed
    snprintf(retryCounterStr, sizeof(retryCounterStr), "%d", panicRetryCounter);

    // Create the message with the retry counter value
    char message[100]; // Adjust the size as needed
    strcpy(message, "Panic has been triggered by the program! Retry counter: ");
    strcat(message, retryCounterStr);

    // Print the message using esp_rom_printf
    esp_rom_printf(message);

    // Call the original panic handler function to finish processing this error (creating a core dump for example...)
    __real_esp_panic_handler(info);
}

const char *reset_reason_to_string(esp_reset_reason_t reason)
{
    switch (reason)
    {
    case ESP_RST_UNKNOWN:
        return "Unknown";
    case ESP_RST_POWERON:
        return "Power-on";
    case ESP_RST_EXT:
        return "External pin";
    case ESP_RST_SW:
        return "Software reset";
    case ESP_RST_PANIC:
        return "Panic reset";
    case ESP_RST_INT_WDT:
        return "Interrupt watchdog reset";
    case ESP_RST_TASK_WDT:
        return "Task watchdog reset";
    case ESP_RST_WDT:
        return "Other watchdog reset";
    case ESP_RST_DEEPSLEEP:
        return "Deep sleep reset";
    case ESP_RST_BROWNOUT:
        return "Brownout reset";
    case ESP_RST_SDIO:
        return "SDIO reset";
    default:
        return "Invalid reset reason value";
    }
}

const char *getChipModelString(esp_chip_model_t chip_model)
{
    switch (chip_model)
    {
    case CHIP_ESP32:
        return "ESP32";
    case CHIP_ESP32S2:
        return "ESP32-S2";
    case CHIP_ESP32S3:
        return "ESP32-S3";
    case CHIP_ESP32C3:
        return "ESP32-C3";
    case CHIP_ESP32C2:
        return "ESP32-C2";
    case CHIP_ESP32C6:
        return "ESP32-C6";
    case CHIP_ESP32H2:
        return "ESP32-H2";
    case CHIP_POSIX_LINUX:
        return "POSIX/Linux Simulator";
    default:
        return "Unknown";
    }
}